import calendar
from datetime import datetime
import re

import aniso8601
from flask import current_app
import six
from werkzeug.utils import cached_property

from flask_potion.utils import get_value
from flask_potion.reference import ResourceReference, ResourceBound
from flask_potion.schema import Schema

class Raw(Schema):
    """
    This is the base class for all field types, can be given any JSON-schema.

    >>> f = fields.Raw({"type": "string"}, io="r")
    >>> f.response
    {'readOnly': True, 'type': 'string'}

    :param io: one of "r", "w" or "rw" (default); used to control presence in fieldsets/parent schemas
    :param schema: JSON-schema for field, or :class:`callable` resolving to a JSON-schema when called
    :param default: optional default value, must be JSON-convertible; may be a callable with no arguments
    :param attribute: key on parent object, optional.
    :param nullable: whether the field is nullable.
    :param title: optional title for JSON schema
    :param description: optional description for JSON schema
    """

    def __init__(self, schema, io="rw", default=None, attribute=None, nullable=False, title=None, description=None):
        self._schema = schema
        self._default = default
        self.attribute = attribute
        self.nullable = nullable
        self.title = title
        self.description = description
        self.io = io

    def _finalize_schema(self, schema, io):
        """
        :return: new schema updated for field `nullable`, `title`, `description` and `default` attributes.
        """
        schema = dict(schema)
        
        if self.io == "r" and "r" in io:
            schema["readOnly"] = True

        if "null" in schema.get("type", []):
            self.nullable = True
        elif self.nullable:
            # enum is independent of type validation:
            if "enum" in schema and None not in schema["enum"]:
                schema["enum"].append(None)

            if "type" in schema:
                type_ = schema["type"]
                if isinstance(type_, (str, dict)):
                    schema["type"] = [type_, "null"]
                else:
                    schema["type"].append("null")

            if "anyOf" in schema:
                if not any("null" in choice.get("type", []) for choice in schema["anyOf"]):
                    schema["anyOf"].append({"type": "null"})
            elif "oneOf" in schema:
                if not any("null" in choice.get("type", []) for choice in schema["oneOf"]):
                    schema["oneOf"].append({"type": "null"})
            elif "type" not in schema:
                if len(schema) == 1 and "$ref" in schema:
                    schema = {"anyOf": [schema, {"type": "null"}]}
                else:
                    current_app.logger.warn('{} is nullable but "null" type cannot be added'.format(self))

        for attr in ("default", "title", "description"):
            value = getattr(self, attr)
            if value is not None:
                schema[attr] = value
        return schema

    @property
    def default(self):
        if callable(self._default):
            return self._default()
        return self._default

    @default.setter
    def default(self, value):
        self._default = value

    def schema(self):
        """
        JSON schema representation
        """
        schema = self._schema
        if callable(schema):
            schema = schema()

        if isinstance(schema, Schema):
            read_schema, write_schema = schema.response, schema.request
        elif isinstance(schema, tuple):
            read_schema, write_schema = schema
        else:
            return self._finalize_schema(schema, "rw")

        return self._finalize_schema(read_schema, "r"), self._finalize_schema(write_schema, "w")

    def format(self, value):
        """
        Format a Python value representation for output in JSON. Noop by default.
        """
        if value is not None:
            return self.formatter(value)
        return value

    def convert(self, instance, validate=True):
        """
        Convert a JSON value representation to a Python object. Noop by default.
        """
        if validate:
            instance = super(Raw, self).convert(instance)

        if instance is not None:
            return self.converter(instance)
        return instance

    def formatter(self, value):
        return value

    def converter(self, value):
        return value

    def output(self, key, obj):
        key = key if self.attribute is None else self.attribute
        return self.format(get_value(key, obj, self.default))

    def __repr__(self):
        return '{}(attribute={})'.format(self.__class__.__name__, repr(self.attribute))


class Any(Raw):
    """
    A field type that allows any value.
    """
    def __init__(self, **kwargs):
        super(Any, self).__init__({"type": ["null", "string", "number", "boolean", "object", "array"]}, **kwargs)

def _field_from_object(parent, cls_or_instance):
    # --- start of Flask-RESTful code ---
    # Copyright (c) 2013, Twilio, Inc.
    # All rights reserved.
    # This code is part of or substantially similar to code in Flask-RESTful and is governed by its
    # license. Please see the LICENSE file in the root of this package.
    if isinstance(cls_or_instance, type):
        container = cls_or_instance()
    else:
        container = cls_or_instance
    if not isinstance(container, Schema):
        raise RuntimeError('{} expected Raw or Schema, but got {}'.format(parent, container.__class__.__name__))
    if not isinstance(container, Raw):
        container = Raw(container)
    # --- end of Flask-RESTful code ---
    return container


class Custom(Raw):
    """
    A field type that cann be passed any schema and optional formatter/converter transformers. It is a very thin
    wrapper over :class:`Raw`.

    :param dict schema: JSON-schema
    :param callable converter: convert function
    :param callable formatter: format function
    """

    def __init__(self, schema, converter=None, formatter=None, **kwargs):
        super(Custom, self).__init__(schema, **kwargs)
        self._converter = converter
        self._formatter = formatter

    def format(self, value):
        if self._formatter is None:
            return value
        return self._formatter(value)

    def converter(self, value):
        if self._converter is None:
            return value
        return self._converter(value)


class Array(Raw, ResourceBound):
    """
    A field for an array of a given field type.

    :param Raw cls_or_instance: field class or instance
    :param int min_items: minimum number of items
    :param int max_items: maximum number of items
    :param bool unique: if ``True``, all values in the list must be unique
    """

    def __init__(self, cls_or_instance, min_items=None, max_items=None, unique=None, **kwargs):
        self.container = container = _field_from_object(self, cls_or_instance)

        schema_properties = [('type', 'array')]
        schema_properties += [(k, v) for k, v in [('minItems', min_items), ('maxItems', max_items), ('uniqueItems', unique)] if v is not None]
        schema = lambda s: dict([('items', s)] + schema_properties)

        super(Array, self).__init__(lambda: (schema(container.response), schema(container.request)),
                                    default=kwargs.pop('default', list), **kwargs)

    def bind(self, resource):
        if isinstance(self.container, ResourceBound):
            self.container = self.container.bind(resource)
        return self

    def format(self, value):
        if value is not None:
            return self.formatter(value)
        if not self.nullable:
            return []
        return value

    def formatter(self, value):
        return [self.container.format(v) for v in value]

    def converter(self, value):
        return [self.container.convert(v) for v in value]


List = Array


class Object(Raw):
    """
    A versatile field for an object, containing either properties all of a single type, properties matching a pattern,
    or named properties matching some fields.

    `Raw.attribute` is not used in pattern properties and additional properties.

    :param properties: field class, instance, or dictionary of {property: field} pairs
    :param str pattern: an optional regular expression that all property keys must match
    :param dict pattern_properties: dictionary of {property: field} pairs
    :param Raw additional_properties: field class or instance
    """

    def __init__(self, properties=None, pattern=None, pattern_properties=None, additional_properties=None, **kwargs):
        self.properties = None
        self.pattern_properties = None
        self.additional_properties = None

        if isinstance(properties, dict):
            self.properties = properties
        elif isinstance(properties, (type, Raw)):
            field = _field_from_object(self, properties)
            if pattern:
                self.pattern_properties = {pattern: field}
            else:
                self.additional_properties = field

        if isinstance(additional_properties, (type, Raw)):
            self.additional_properties = _field_from_object(self, additional_properties)
        elif additional_properties is True:
            self.additional_properties = Any()

        if isinstance(pattern_properties, (type, Raw)):
            self.pattern_properties = _field_from_object(self, pattern_properties)

        def schema():
            request = {"type": "object"}
            response = {"type": "object"}

            for schema, attr in ((request, "request"), (response, "response")):
                if self.properties:
                    schema["properties"] = {key: getattr(field, attr) for key, field in self.properties.items()}
                if self.pattern_properties:
                    schema["patternProperties"] = {pattern: getattr(field, attr)
                                                   for pattern, field in self.pattern_properties.items()}
                if self.additional_properties:
                    schema["additionalProperties"] = getattr(self.additional_properties, attr)
                else:
                    schema["additionalProperties"] = False

            return response, request

        if self.pattern_properties and (len(self.pattern_properties) > 1 or self.additional_properties):
            raise NotImplementedError("Only one pattern property is currently supported "
                                      "and it cannot be combined with additionalProperties")

        super(Object, self).__init__(schema, **kwargs)

    @cached_property
    def _property_attributes(self):
        if not self.properties:
            return ()
        return [field.attribute or key for key, field in self.properties.items()]

    def formatter(self, value):
        output = {}

        if self.properties:
            output = {key: field.format(get_value(field.attribute or key, value, field.default)) for key, field in self.properties.items()}
        else:
            output = {}

        if self.pattern_properties:
            pattern, field = next(iter(self.pattern_properties.items()))

            if not self.additional_properties:
                output.update({k: field.format(v) for k, v in value.items() if k not in self._property_attributes})
            else:
                raise NotImplementedError()
                # TODO match regular expression
        elif self.additional_properties:
            field = self.additional_properties
            output.update({k: field.format(v) for k, v in value.items() if k not in self._property_attributes})

        return output

    def converter(self, instance):
        result = {}

        if self.properties:
            result = {field.attribute or key: field.convert(instance.get(key, field.default))
                      for key, field in self.properties.items()}

        if self.pattern_properties:
            pattern, field = next(iter(self.pattern_properties.items()))

            if not self.additional_properties:
                result.update({key: field.convert(value)
                               for key, value in instance.items() if key not in result})
            else:
                raise NotImplementedError()
                # TODO match regular expression
        elif self.additional_properties:
            field = self.additional_properties
            result.update({key: field.convert(value) for key, value in instance.items() if key not in result})

        return result


class AttributeMapped(Object):
    """
    Maps property keys from a JSON object to a list of items using `mapping_attribute`. The mapping attribute is the
    name of the attribute where the value of the property key is set on the property values.

    :class:`sa.InlineModel` is typically used with this field in a common SQLAlchemy pattern.

    :param Raw cls_or_instance: field class or instance
    :param str pattern: an optional regular expression that all property keys must match
    :param str mapping_attribute: mapping attribute
    """

    def __init__(self, cls_or_instance, mapping_attribute=None, **kwargs):
        self.mapping_attribute = mapping_attribute
        # TODO reject additional_properties, properties, pattern_properties, pattern
        super(AttributeMapped, self).__init__(cls_or_instance, **kwargs)

    def _set_mapping_attribute(self, obj, value):
        if isinstance(obj, dict):
            obj[self.mapping_attribute] = value
        else:
            setattr(obj, self.mapping_attribute, value)
        return obj

    def formatter(self, value):
        if self.pattern_properties:
            pattern, field = next(iter(self.pattern_properties.items()))
            return {get_value(self.mapping_attribute, v, None): field.format(v) for v in value}
        elif self.additional_properties:
            return {get_value(self.mapping_attribute, v, None): self.additional_properties.format(v) for v in value}

    def converter(self, value):
        if self.pattern_properties:
            pattern, field = next(iter(self.pattern_properties.items()))
            return [self._set_mapping_attribute(field.convert(v), k) for k, v in value.items()]
        elif self.additional_properties:
            return [self._set_mapping_attribute(self.additional_properties.convert(v), k) for k, v in value.items()]


class String(Raw):
    """
    :param int min_length: minimum length of string
    :param int max_length: maximum length of string
    :param str pattern: regex pattern that the string must match
    :param list enum: list of strings with enumeration
    """

    def __init__(self, min_length=None, max_length=None, pattern=None, enum=None, format=None, **kwargs):
        schema = {"type": "string"}

        if enum is not None:
            enum = list(enum)

        for v, k in ((min_length, 'minLength'),
                     (max_length, 'maxLength'),
                     (pattern, 'pattern'),
                     (enum, 'enum'),
                     (format, 'format')):
            if v is not None:
                schema[k] = v

        super(String, self).__init__(schema, **kwargs)

try:
    from datetime import timezone
except ImportError:
    from datetime import tzinfo, timedelta

    class timezone(tzinfo):
        def __init__(self, utcoffset, name=None):
            self._utcoffset = utcoffset
            self._name = name

        def utcoffset(self, dt):
            return self._utcoffset

        def tzname(self, dt):
            return self._name

        def dst(self, dt):
            return timedelta(0)

    timezone.utc = timezone(timedelta(0), 'UTC')


class Date(Raw):
    """
    A field for EJSON-style dates in the format:

    ::

        {"$date": MILLISECONDS_SINCE_EPOCH}

    Converts to :class:`datetime.date` with UTC timezone.

    """

    def __init__(self, **kwargs):
        # TODO is a 'format' required for "date"
        super(Date, self).__init__({
                                       "type": "object",
                                       "properties": {
                                           "$date": {
                                               "type": "integer"
                                           }
                                       },
                                       "additionalProperties": False
                                   }, **kwargs)

    def formatter(self, value):
        return {"$date": int(calendar.timegm(value.timetuple()) * 1000)}

    def converter(self, value):
        # TODO support both $dateObj and ISO string formats
        return datetime.fromtimestamp(value["$date"] / 1000, timezone.utc).date()


class DateTime(Date):
    """
    A field for EJSON-style date-times in the format:

    ::

        {"$date": MILLISECONDS_SINCE_EPOCH}

    Converts to :class:`datetime.datetime` with UTC timezone.

    """

    def formatter(self, value):
        return {"$date": int(calendar.timegm(value.utctimetuple()) * 1000)}

    def converter(self, value):
        # TODO support both $dateObj and ISO string formats
        return datetime.fromtimestamp(value["$date"] / 1000, timezone.utc)


class DateString(Raw):
    """
    A field for ISO8601-formatted date strings.
    """

    def __init__(self, **kwargs):
        # TODO is a 'format' required for "date"
        super(DateString, self).__init__({"type": "string", "format": "date"}, **kwargs)

    def formatter(self, value):
        return value.strftime('%Y-%m-%d')

    def converter(self, value):
        return aniso8601.parse_date(value)


class DateTimeString(Raw):
    """
    A field for ISO8601-formatted date-time strings.
    """

    def __init__(self, **kwargs):
        super(DateTimeString, self).__init__({"type": "string", "format": "date-time"}, **kwargs)

    def formatter(self, value):
        return value.isoformat()

    def converter(self, value):
        # FIXME enforce UTC
        return aniso8601.parse_datetime(value)


class Uri(String):
    def __init__(self, **kwargs):
        super(Uri, self).__init__(format="uri", **kwargs)


class Email(String):
    def __init__(self, **kwargs):
        super(Email, self).__init__(format="email", **kwargs)


class Boolean(Raw):
    def __init__(self, **kwargs):
        super(Boolean, self).__init__({"type": "boolean"}, **kwargs)

    def format(self, value):
        return bool(value)


class Integer(Raw):
    def __init__(self, minimum=None, maximum=None, default=None, **kwargs):
        schema = {"type": "integer"}

        if minimum is not None:
            schema['minimum'] = minimum
        if maximum is not None:
            schema['maximum'] = maximum

        super(Integer, self).__init__(schema, default=default, **kwargs)

    def formatter(self, value):
        return int(value)


class PositiveInteger(Integer):
    """
    A :class:`Integer` field that only accepts integers >=1.
    """

    def __init__(self, maximum=None, **kwargs):
        super(PositiveInteger, self).__init__(minimum=1, maximum=maximum, **kwargs)


class Number(Raw):
    def __init__(self,
                 minimum=None,
                 maximum=None,
                 exclusive_minimum=False,
                 exclusive_maximum=False,
                 **kwargs):

        schema = {"type": "number"}

        if minimum is not None:
            schema['minimum'] = minimum
            if exclusive_minimum:
                schema['exclusiveMinimum'] = True

        if maximum is not None:
            schema['maximum'] = maximum
            if exclusive_maximum:
                schema['exclusiveMaximum'] = True

        super(Number, self).__init__(schema, **kwargs)

    def formatter(self, value):
        return float(value)


class ToOne(Raw, ResourceBound):
    """
    Represents references between resources as `json-ref` objects.

    Resource references can be one of the following:

    - :class:`Resource` class
    - a string with a resource name
    - a string with a module name and class name of a resource
    - ``"self"`` --- which resolves to the resource this field is bound to

    :param resource: a resource reference
    """
    def __init__(self, resource, **kwargs):
        self.target_reference = ResourceReference(resource)

        def schema():
            target = self.target
            key_converters = self.target.meta.key_converters
            response_schema = self.formatter_key.response
            if len(key_converters) > 1:
                request_schema = {
                    # "type": [self.formatter_key.matcher_type()] + [nk.matcher_type() for nk in natural_keys],
                    "anyOf": [nk.request for nk in key_converters]
                }
            else:
                request_schema = self.formatter_key.request
            return response_schema, request_schema

        super(ToOne, self).__init__(schema, **kwargs)

    def rebind(self, resource):
        if self.target_reference.value == 'self':
            return self.__class__(
                'self',
                default=self.default,
                attribute=self.attribute,
                nullable=self.nullable,
                title=self.title,
                description=self.description,
                io=self.io
            ).bind(resource)
        else:
            return self

    @cached_property
    def target(self):
        return self.target_reference.resolve(self.resource)

    @cached_property
    def formatter_key(self):
        return self.target.meta.key_converters[0]

    def formatter(self, item):
        return self.formatter_key.format(item)

    def converter(self, value):
        for python_type, json_type in (
                (dict, 'object'),
                (int, 'integer'),
                ((list, tuple), 'array'),
                (six.string_types, 'string')):
            if isinstance(value, python_type):
                return self.target.meta.key_converters_by_type[json_type].convert(value)


class ToMany(Array):
    def __init__(self, resource, **kwargs):
        super(ToMany, self).__init__(ToOne(resource, nullable=False), **kwargs)


class Inline(Raw, ResourceBound):
    """
    Formats and converts items in a :class:`ModelResource` using the resource's ``schema``.

    :param resource: a resource reference as in :class:`ToOne`
    :param bool patch_instance: whether to allow partial objects
    """

    def __init__(self, resource, patchable=False, **kwargs):
        self.target_reference = ResourceReference(resource)
        self.patchable = patchable

        def schema():
            def _response_schema():
                if self.resource == self.target:
                    return {"$ref": "#"}
                return {"$ref": self.resource.routes["schema"].rule_factory(self.resource)}

            if not not self.patchable:
                return _response_schema()
            else:
                return _response_schema(), self.target.schema.patchable.request

        super(Inline, self).__init__(schema, **kwargs)

    def rebind(self, resource):
        if self.target_reference.value == 'self':
            return self.__class__(
                'self',
                patchable=self.patchable,
                default=self.default,
                attribute=self.attribute,
                nullable=self.nullable,
                title=self.title,
                description=self.description,
                io=self.io
            ).bind(resource)
        else:
            return self

    @cached_property
    def target(self):
        return self.target_reference.resolve(self.resource)

    def format(self, item):
        return self.target.schema.format(item)

    def convert(self, item):
        return self.target.schema.convert(item, patchable=self.patchable)


class ItemType(Raw):
    """
    A string field that formats the name of a resource; read-only.
    """
    def __init__(self, resource):
        self.resource = resource
        super(ItemType, self).__init__(lambda: {
            "type": "string",
            "enum": [self.resource.meta.name]
        }, io="r")

    def format(self, value):
        return self.resource.meta.name


class ItemUri(Raw):
    """
    A string field that formats the url of a resource item; read-only.
    """
    def __init__(self, resource, attribute=None):
        self.resource = resource
        super(ItemUri, self).__init__(lambda: {
            "type": "string",
            "pattern": "^{}\/[^/]+$".format(re.escape(resource.route_prefix))
        }, io="r", attribute=attribute)

    def format(self, value):
        return '{}/{}'.format(self.resource.route_prefix, value)


class sa:
    class InlineModel(Object):
        """
        :param dict properties:
        :param model:
        """
        def __init__(self, properties, model, **kwargs):
            super(sa.InlineModel, self).__init__(properties, **kwargs)
            self.model = model

        def converter(self, instance):
            instance = super(sa.InlineModel, self).converter(instance)
            # obj = EmbeddedJob.complete(super().convert(obj))
            # TODO commit()?
            if instance is not None:
                instance = self.model(**instance)
            return instance



from __future__ import absolute_import

from bson import ObjectId as bson_ObjectId
from flask import current_app
from mongoengine.errors import OperationError, ValidationError
from bson.errors import InvalidId

import mongoengine.fields as mongo_fields

from flask_potion.utils import get_value
from flask_potion.exceptions import ItemNotFound, BackendConflict
from flask_potion.backends import Manager, Pagination
from flask_potion.signals import before_create, before_update, after_update, before_delete, after_delete, after_create, \
    before_add_to_relation, after_remove_from_relation, before_remove_from_relation, after_add_to_relation
from flask_potion import fields


# TODO: more elaborate field that validates and returns ObjectId

class custom_fields:
    class ObjectId(fields.String):
        def formatter(self, value):
            if isinstance(value, bson_ObjectId):
                return str(value)
            else:
                return value


MONGO_REFERENCE_FIELD_TYPES = (
    mongo_fields.ReferenceField,
    mongo_fields.CachedReferenceField,
    mongo_fields.ObjectIdField
)


COMPARATOR_EXPRESSIONS = {
    '$eq': lambda column, value: {column: value},
    '$ne': lambda column, value: {"%s__ne" % column: value},
    '$in': lambda column, value: {"%s__in" % column: value},
    '$lt': lambda column, value: {"%s__lt" % column: value},
    '$gt': lambda column, value: {"%s__gt" % column: value},
    '$lte': lambda column, value: {"%s__lte" % column: value},
    '$gte': lambda column, value: {"%s__gte" % column: value},
    '$contains': lambda column, value: {"%s__contains" % column: value},
    '$startswith': lambda column, value: {"%s__startswith" % column: value.replace('%', '\\%')},
    '$endswith': lambda column, value: {"%s__endswith" % column: value.replace('%', '\\%')},
    # TODO: $istartswith and $iendswith filters
}

MONGO_FIELDS_MAPPING = {
    mongo_fields.ReferenceField: fields.Object,
    mongo_fields.ObjectIdField: custom_fields.ObjectId,
    mongo_fields.IntField: fields.Integer,
    mongo_fields.FloatField: fields.Number,
    mongo_fields.BooleanField: fields.Boolean,
    mongo_fields.ListField: fields.Array,
    mongo_fields.SortedListField: fields.Array,
    mongo_fields.LongField: fields.Number,
    mongo_fields.BinaryField: fields.Array,
    mongo_fields.DictField: fields.Object,
    mongo_fields.DateTimeField: fields.Date,
    mongo_fields.ComplexDateTimeField: fields.DateTime
}


class MongoEngineManager(Manager):
    """
    A manager for MongoEngineManager models.

    Expects that :class:`Meta.model` contains an MongoEngine declarative model.

    """
    supported_comparators = tuple(COMPARATOR_EXPRESSIONS.keys())

    def __init__(self, resource, model):
        super(MongoEngineManager, self).__init__(resource, model)
        meta = resource.meta

        self.id_attribute = meta.id_attribute
        self.id_column = model._fields[self.id_attribute]

        if resource.meta.include_id:
            resource.schema.set("$id", custom_fields.ObjectId(io='r', attribute=self.id_attribute))

        # resource name: use model table's name if not set explicitly
        if not hasattr(resource.Meta, 'name'):
            meta['name'] = model._meta.get('collection', model.__class__.__name__).lower()

        fs = resource.schema
        include_fields = meta.get('include_fields', None)
        exclude_fields = meta.get('exclude_fields', None)
        read_only_fields = meta.get('read_only_fields', ())
        write_only_fields = meta.get('write_only_fields', ())
        pre_declared_fields = {f.attribute or k for k, f in fs.fields.items()}

        for name, column in model._fields.items():
            if (include_fields and name in include_fields) or \
                    (exclude_fields and name not in exclude_fields) or \
                    not (include_fields or exclude_fields):

                if column.primary_key:
                    continue
                if name in pre_declared_fields:
                    continue
                if isinstance(column, MONGO_REFERENCE_FIELD_TYPES):
                    continue
                if isinstance(column, mongo_fields.ListField) and isinstance(column.field, MONGO_REFERENCE_FIELD_TYPES):
                    continue

                field_class, args, kwargs = self._get_field_from_mongoengine_type(column)

                io = "rw"
                if name in read_only_fields:
                    io = "r"
                elif name in write_only_fields:
                    io = "w"

                if not (column.null or column.default is not None):
                    fs.required.add(name)

                fs.set(name, field_class(*args, io=io, attribute=name, **kwargs))

    def _init_key_converters(self, resource, meta):
        meta.id_attribute = meta.get('id_attribute', meta.model.pk)
        meta.id_field_class = custom_fields.ObjectId
        meta.id_converter = 'string'
        super(MongoEngineManager, self)._init_key_converters(resource, meta)

    def _get_field_from_mongoengine_type(self, property):
        args = ()
        kwargs = {}

        if isinstance(property, mongo_fields.ListField):
            field_class = fields.Array
            args = (self._get_field_from_mongoengine_type(property.field)[0],)
        elif isinstance(property, mongo_fields.UUIDField):  # TODO support UUIDfield
            field_class = fields.String
            kwargs = {
                'max_length': 36,
                'min_length': 36,
            }
        elif isinstance(property, mongo_fields.StringField):
            field_class = fields.String
            kwargs = {
                'max_length': property.max_length,
                'min_length': property.min_length,
                'enum': property.choices,
                'pattern': property.regex
            }
        else:
            try:
                field_class = MONGO_FIELDS_MAPPING[type(property)]
            except KeyError:
                raise TypeError("%s not supported (use %s)" % (type(property), str(list(MONGO_FIELDS_MAPPING.keys()))))

        kwargs['nullable'] = property.null
        kwargs['default'] = property.default
        return field_class, args, kwargs

    def _where_expression(self, where):
        expressions = {}

        for condition in where:
            expressions.update(COMPARATOR_EXPRESSIONS[condition.comparator.name](condition.attribute, condition.value))

        return expressions

    @staticmethod
    def _order_by(sort):
        for attribute, reverse in sort:
            if reverse:
                yield "-%s" % attribute
            else:
                yield "%s" % attribute

    def relation_instances(self, item, attribute, target_resource, page=None, per_page=None):
        query = getattr(item, attribute)
        if page and per_page:
            return Pagination.from_list(query, page, per_page)
        else:
            return query.all()

    def relation_add(self, item, attribute, target_resource, target_item):
        before_add_to_relation.send(self.resource, item=item, attribute=attribute, child=target_item)
        getattr(item, attribute).append(target_item)
        item.save()
        after_add_to_relation.send(self.resource, item=item, attribute=attribute, child=target_item)

    def relation_remove(self, item, attribute, target_resource, target_item):
        before_remove_from_relation.send(self.resource, item=item, attribute=attribute, child=target_item)
        getattr(item, attribute).remove(target_item)
        item.save()
        after_remove_from_relation.send(self.resource, item=item, attribute=attribute, child=target_item)

    def paginated_instances(self, page, per_page, where=None, sort=None):
        instances = self.instances(where=where, sort=sort)
        return Pagination(instances, page, per_page, instances.count())

    def instances(self, where=None, sort=None):
        query = self.model.objects

        if where is not None:
            where_expression = self._where_expression(where)
            query = query(**where_expression)

        if sort is not None:
            query = query.order_by(*self._order_by(sort))

        return query

    def first(self, where=None, sort=None):
        res = self.instances(where, sort).first()
        if res is None:
            raise ItemNotFound(self.resource, where=where)
        else:
            return res

    def create(self, properties, commit=True):
        item = self.model()

        for key, value in properties.items():
            setattr(item, key, value)

        before_create.send(self.resource, item=item)

        try:
            item.save()
        except OperationError as e:
            if current_app.debug:
                raise BackendConflict(debug_info=dict(statement=e.args))
            raise BackendConflict()

        after_create.send(self.resource, item=item)
        return item

    def read(self, id):
        try:
            return self.model.objects(**{self.id_attribute: id}).first()
        except (InvalidId, ValidationError):
            raise ItemNotFound(self.resource, id=id)

    def update(self, item, changes, commit=True):
        actual_changes = {
            key: value for key, value in changes.items()
            if get_value(key, item, None) != value
        }

        try:
            before_update.send(self.resource, item=item, changes=actual_changes)

            for key, value in changes.items():
                setattr(item, key, value)

            if commit:
                item.save()
        except OperationError:
            raise

        after_update.send(self.resource, item=item, changes=actual_changes)
        return item

    def delete(self, item):
        before_delete.send(self.resource, item=item)
        item.delete()
        after_delete.send(self.resource, item=item)



from collections import OrderedDict
import inspect
import operator

from flask import current_app, make_response, json
from six import wraps
from werkzeug.wrappers import BaseResponse

from .exceptions import PotionException
from .routes import RouteSet
from .utils import unpack
from .resource import Resource, ModelResource

__all__ = (
    'Api',
    'Resource',
    'ModelResource',
    'fields',
    'routes',
    'schema',
    'signals',
    'contrib',
    'natural_keys'
)


class Api(object):
    """
    This is the Potion extension.

    You need to register :class:`Api` with a :class:`Flask` application either upon initializing :class:`Api` or later using :meth:`init_app()`.

    :param app: a :class:`Flask` instance
    :param list decorators: an optional list of decorator functions
    :param prefix: an optional API prefix. Must start with "/"
    :param str title: an optional title for the schema
    :param str description: an optional description for the schema
    :param Manager default_manager: an optional manager to use as default. If SQLAlchemy is installed, will use :class:`backends.alchemy.SQLAlchemyManager`
    """
    def __init__(self, app=None, decorators=None, prefix=None, title=None, description=None, default_manager=None):
        self.app = app
        self.prefix = prefix or ''
        self.decorators = decorators or []
        self.title = title
        self.description = description
        self.endpoints = set()
        self.resources = {}
        self.views = []

        self.default_manager = None
        if default_manager is None:
            try:
                from .backends.alchemy import SQLAlchemyManager
                self.default_manager = SQLAlchemyManager
            except ImportError:
                pass
        else:
            self.default_manager = default_manager

        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        """
        :param app: a :class:`Flask` instance
        """
        app.potion = self  # NOTE in part because of this, only one Api instance can exist per app
        app.config.setdefault('POTION_MAX_PER_PAGE', 100)
        app.config.setdefault('POTION_DEFAULT_PER_PAGE', 20)

        app.add_url_rule(
            rule=''.join((self.prefix, '/schema')),
            view_func=self.output(self._schema_view),
            endpoint='schema',
            methods=['GET'])

        for rule, view, endpoint, methods in self.views:
            app.add_url_rule(rule, view_func=view, endpoint=endpoint, methods=methods)

        @app.errorhandler(PotionException)
        def handle_invalid_usage(error):
            return error.make_response()

    def output(self, view):
        @wraps(view)
        def wrapper(*args, **kwargs):
            resp = view(*args, **kwargs)
            if isinstance(resp, BaseResponse):
                return resp

            data, code, headers = unpack(resp)

            settings = {}
            if current_app.debug:
                settings.setdefault('indent', 4)
                settings.setdefault('sort_keys', True)

            data = json.dumps(data, **settings)
            resp = make_response(data, code)
            resp.headers.extend(headers or {})
            resp.headers['Content-Type'] = 'application/json'
            return resp

        return wrapper

    def _schema_view(self):
        schema = OrderedDict()
        schema["$schema"] = "http://json-schema.org/draft-04/hyper-schema#"

        if self.title:
            schema["title"] = self.title
        if self.description:
            schema["description"] = self.description

        # schema["definitions"] = definitions = OrderedDict([])
        schema["properties"] = properties = OrderedDict([])
        for name, resource in sorted(self.resources.items(), key=operator.itemgetter(0)):
            resource_schema_rule = resource.routes['schema'].rule_factory(resource)
            properties[name] = {"$ref": '{}#'.format(resource_schema_rule)}

        return OrderedDict(schema), 200, {'Content-Type': 'application/schema+json'}

    def add_route(self, route, resource, endpoint=None):
        endpoint = endpoint or '.'.join((resource.meta.name, route.attribute))
        methods = route.methods()
        rule = route.rule_factory(resource)
        view = self.output(route.view_factory(endpoint, resource))

        for decorator in self.decorators:
            view = decorator(view)

        if self.app:
            self.app.add_url_rule(rule, view_func=view, endpoint=endpoint, methods=methods)
        else:
            self.views.append((rule, view, endpoint, methods))

    def add_resource(self, resource):
        """
        Add a :class:`Resource` class to the API and generate endpoints for all its routes.

        :param Resource resource: resource
        :return:
        """
        # prevent resources from being added twice
        if resource in self.resources.values():
            return

        # check that each model resource has a manager; if not, initialize it.
        if issubclass(resource, ModelResource) and resource.manager is None:
            if self.default_manager:
                resource.manager = self.default_manager(resource, resource.meta.get('model'))
            else:
                raise RuntimeError("'{}' has no manager, and no default manager has been defined. "
                                   "If you're using Potion with SQLAlchemy, "
                                   "ensure you have installed Flask-SQLAlchemy.".format(resource.meta.name))

        resource.api = self
        resource.route_prefix = ''.join((self.prefix, '/', resource.meta.name))

        for route in resource.routes.values():
            self.add_route(route, resource)

        for name, rset in inspect.getmembers(resource, lambda m: isinstance(m, RouteSet)):
            if rset.attribute is None:
                rset.attribute = name

            for i, route in enumerate(rset.routes()):
                if route.attribute is None:
                    route.attribute = '{}_{}'.format(rset.attribute, i)
                resource.routes['{}_{}'.format(rset.attribute, route.attribute)] = route
                self.add_route(route, resource)

        self.resources[resource.meta.name] = resource



from __future__ import division
import datetime
from math import ceil

import six
from flask_potion.exceptions import ItemNotFound
from flask_potion import fields


class Manager(object):
    """

    .. attribute:: supported_comparators

        A tuple of names filter comparators supported by this manager.

    :param flask_potion.resource.Resource resource: resource class
    :param model: model read from ``Meta.model`` or ``None``
    """
    supported_comparators = ()

    def __init__(self, resource, model):
        self.resource = resource
        self.model = model

        self._init_key_converters(resource, resource.meta)

    def _init_key_converters(self, resource, meta):
        if 'natural_key' in meta:
            from flask_potion.natural_keys import PropertyKey, PropertiesKey
            if isinstance(meta.natural_key, str):
                meta['key_converters'] += (PropertyKey(meta.natural_key),)
            elif isinstance(meta.natural_key, (list, tuple)):
                meta['key_converters'] += (PropertiesKey(*meta.natural_key),)

        if 'key_converters' in meta:
            meta.key_converters = [k.bind(resource) for k in meta['key_converters']]
            meta.key_converters_by_type = {}
            for nk in meta.key_converters:
                if nk.matcher_type() in meta.key_converters_by_type:
                    raise RuntimeError(
                        'Multiple keys of type {} defined for {}'.format(nk.matcher_type(), meta.name))
                meta.key_converters_by_type[nk.matcher_type()] = nk

    @staticmethod
    def _get_field_from_python_type(python_type):
        try:
            return {
                str: fields.String,
                six.text_type: fields.String,
                int: fields.Integer,
                float: fields.Number,
                bool: fields.Boolean,
                list: fields.Array,
                dict: fields.Object,
                datetime.date: fields.Date,
                datetime.datetime: fields.DateTime
            }[python_type]
        except KeyError:
            raise RuntimeError('No appropriate field class for "{}" type found'.format(python_type))

    def relation_instances(self, item, attribute, target_resource, page=None, per_page=None):
        """

        :param item:
        :param attribute:
        :param target_resource:
        :param page:
        :param per_page:
        :return:
        """
        raise NotImplementedError()

    def relation_add(self, item, attribute, target_resource, target_item):
        """

        :param item:
        :param attribute:
        :param target_resource:
        :param target_item:
        :return:
        """
        raise NotImplementedError()

    def relation_remove(self, item, attribute, target_resource, target_item):
        """

        :param item:
        :param attribute:
        :param target_resource:
        :param target_item:
        :return:
        """
        raise NotImplementedError()

    def paginated_instances(self, page, per_page, where=None, sort=None):
        """

        :param page:
        :param per_page:
        :param where:
        :param sort:
        :return: a :class:`Pagination` object or similar
        """
        pass

    def instances(self, where=None, sort=None):
        """

        :param where:
        :param sort:
        :return:
        """
        pass

    def first(self, where=None, sort=None):
        """

        :param where:
        :param sort:
        :return:
        :raises exceptions.ItemNotFound:
        """
        try:
            return self.instances(where, sort)[0]
        except IndexError:
            raise ItemNotFound(self.resource, where=where)

    def create(self, properties, commit=True):
        """

        :param properties:
        :param commit:
        :return:
        """
        pass

    def read(self, id):
        """

        :param id:
        :return:
        """
        pass

    def update(self, item, changes, commit=True):
        """

        :param item:
        :param changes:
        :param commit:
        :return:
        """
        pass

    def delete(self, item):
        """

        :param item:
        :return:
        """
        pass

    def delete_by_id(self, id):
        """

        :param id:
        :return:
        """
        return self.delete(self.read(id))

    def commit(self):
        pass

    def begin(self):

        pass


class Pagination(object):
    """
    A pagination class for list-like instances.

    :param items:
    :param page:
    :param per_page:
    :param total:
    """

    def __init__(self, items, page, per_page, total):
        self.items = items
        self.page = page
        self.per_page = per_page
        self.total = total

    @property
    def pages(self):
        return max(1, int(ceil(self.total / self.per_page)))

    @property
    def has_prev(self):
        return self.page > 1

    @property
    def has_next(self):
        return self.page < self.pages

    @classmethod
    def from_list(cls, items, page, per_page):
        start = per_page * (page - 1)
        return Pagination(items[start:start + per_page], page, per_page, len(items))


from collections import OrderedDict
import inspect
from operator import attrgetter
import itertools

import six

from .natural_keys import RefKey, IDKey, PropertyKey, PropertiesKey
from .fields import ItemType, ItemUri, Integer, Inline
from .reference import ResourceBound
from .instances import Instances
from .utils import AttributeDict
from .routes import Route
from .schema import FieldSet


class ResourceMeta(type):
    def __new__(mcs, name, bases, members):
        class_ = super(ResourceMeta, mcs).__new__(mcs, name, bases, members)
        class_.routes = routes = dict(getattr(class_, 'routes') or {})
        class_.meta = meta = AttributeDict(getattr(class_, 'meta', {}) or {})

        for base in bases:
            for n, m in inspect.getmembers(base, lambda m: isinstance(m, Route)):
                if m.attribute is None:
                    m.attribute = n
                routes[m.attribute] = m

            if hasattr(base, 'Meta'):
                meta.update(base.Meta.__dict__)


        if 'Meta' in members:
            changes = members['Meta'].__dict__
            for k, v in changes.items():
                if not k.startswith('__'):
                    meta[k] = v

            if not changes.get('name', None):
                meta['name'] = name.lower()
        else:
            meta['name'] = name.lower()

        schema = {}
        for base in bases:
            if hasattr(base, 'Schema'):
                schema.update(base.Schema.__dict__)

        if 'Schema' in members:
            schema.update(members['Schema'].__dict__)

        if schema:
            # TODO support FieldSet with definitions
            class_.schema = fs = FieldSet({k: f for k, f in schema.items() if not k.startswith('__')},
                                          required_fields=meta.get('required_fields', None))

            for name in meta.get('read_only_fields', ()):
                if name in fs.fields:
                    fs.fields[name].io = "r"

            for name in meta.get('write_only_fields', ()):
                if name in fs.fields:
                    fs.fields[name].io = "w"

            fs.bind(class_)

        for n, m in members.items():
            if isinstance(m, Route):
                if m.attribute is None:
                    m.attribute = n

                routes[m.attribute] = m

            if isinstance(m, ResourceBound):
                m.bind(class_)

        return class_


class Resource(six.with_metaclass(ResourceMeta, object)):
    """
    A plain resource with nothing but a schema.


    .. attribute:: api

        Back reference to the :class:`Api` this resource is registered on.

    .. attribute:: meta

        A :class:`AttributeDict` of configuration attributes collected from the :class:`Meta` attributes of the base classes.

    .. attribute:: routes

        A dictionary of routes registered with this resource. Keyed by ``Route.attribute``.

    .. attribute:: schema

        A :class:`FieldSet` containing fields collected from the :class:`Schema` attributes of the base classes.

    .. attribute:: route_prefix

        The prefix URI to any route in this resource; includes the API prefix.

    .. method:: described_by

        A :class:`Route` at ``/schema`` that contains the JSON Hyper-Schema for this resource.

    """
    api = None
    meta = None
    routes = None
    schema = None
    route_prefix = None

    @Route.GET('/schema', rel="describedBy", attribute="schema")
    def described_by(self): # No "targetSchema" because that would be way too meta.
        schema = OrderedDict([
            ("$schema", "http://json-schema.org/draft-04/hyper-schema#"),
        ])

        # copy title, description from ModelResource.meta
        for property in ('title', 'description'):
            value = getattr(self.meta, property)
            if value:
                schema[property] = value

        links = itertools.chain(*(route.links() for name, route in sorted(self.routes.items())))

        if self.schema:
            schema['type'] = "object"
            schema.update(self.schema.response)

        # TODO more intuitive sorting [self, instances,.. GET_bar, POST_bar, GET_foo, POST_foo, ..describedBy]
        schema['links'] = [link.schema_factory(self) for link in sorted(links, key=attrgetter('relation'))]

        return schema, 200, {'Content-Type': 'application/schema+json'}

    class Meta:
        name = None
        title = None
        description = None
        required_fields = None
        read_only_fields = ()
        write_only_fields = ()


class ModelResourceMeta(ResourceMeta):
    def __new__(mcs, name, bases, members):
        class_ = super(ModelResourceMeta, mcs).__new__(mcs, name, bases, members)

        if 'Meta' in members:
            meta = class_.meta
            changes = members['Meta'].__dict__

            if 'model' in changes or 'model' in meta and 'manager' in changes:
                fs = class_.schema

                if meta.include_id:
                    fs.set('$id', meta.id_field_class(io="r", attribute=meta.id_attribute))
                else:
                    fs.set('$uri', ItemUri(class_, attribute=meta.id_attribute))

                if meta.manager is not None:
                    class_.manager = meta.manager(class_, meta.model)

                if meta.include_type:
                    fs.set('$type', ItemType(class_))
        return class_


class ModelResource(six.with_metaclass(ModelResourceMeta, Resource)):
    """

    .. method:: create

        A link --- part of a :class:`Route` at the root of the resource --- for creating new items.

        :param properties:
        :return: created item

    .. method:: instances

        A link --- part of a :class:`Route` at the root of the resource --- for reading item instances.

        :param where:
        :param sort:
        :param int page:
        :param int per_page:
        :return: list of items

    .. method:: read

        A link --- part of a :class:`Route` at ``/<{Resource.meta.id_converter}:id>`` --- for reading a specific item.

        :param id: item id
        :return: item

    .. method:: update

        A link --- part of a :class:`Route` at ``/<{Resource.meta.id_converter}:id>`` --- for updating a specific item.

        :param id: item id
        :param properties: changes
        :return: item

    .. method:: destroy

        A link --- part of a :class:`Route` at ``/<{Resource.meta.id_converter}:id>`` --- for deleting a specific item.

        :param id: item id
        :return: ``(None, 204)``

    """
    manager = None

    @Route.GET('', rel="instances")
    def instances(self, **kwargs):
        return self.manager.paginated_instances(**kwargs)

    # TODO custom schema (Instances/Instances) that contains the necessary schema.
    instances.request_schema = instances.response_schema = Instances() # TODO NOTE Instances('self') for filter, etc. schema


    @instances.POST(rel="create")
    def create(self, properties):  # XXX need some way for field bindings to be dynamic/work dynamically.
        item = self.manager.create(properties)
        return item  # TODO consider 201 Created

    create.request_schema = create.response_schema = Inline('self')

    @Route.GET(lambda r: '/<{}:id>'.format(r.meta.id_converter), rel="self", attribute="instance")
    def read(self, id):
        return self.manager.read(id)

    read.request_schema = None
    read.response_schema = Inline('self')

    @read.PATCH(rel="update")
    def update(self, properties, id):
        item = self.manager.read(id)
        updated_item = self.manager.update(item, properties)
        return updated_item

    update.request_schema = Inline('self', patchable=True)
    update.response_schema = update.request_schema

    @update.DELETE(rel="destroy")
    def destroy(self, id):
        self.manager.delete_by_id(id)
        return None, 204

    class Schema:
        pass

    class Meta:
        id_attribute = 'id'
        id_converter = 'int'
        id_field_class = Integer  # Must inherit from Integer, String or ItemUri
        include_id = False
        include_type = False
        manager = None
        include_fields = None
        exclude_fields = None
        allowed_filters = "*"
        permissions = {
            "read": "anyone",  # anybody, yes
            "create": "nobody",  # noone, no
            "update": "create",
            "delete": "update"
        }
        postgres_text_search_fields = ()
        postgres_full_text_index = None  # $fulltext
        cache = False
        key_converters = (
            RefKey(),
            IDKey()
        )
        natural_key = None


from collections import OrderedDict
import re
from types import MethodType
import sys

from flask import request
from werkzeug.utils import cached_property

from flask_potion.fields import _field_from_object
from flask_potion.fields import ToOne, Integer, Object
from flask_potion.utils import get_value
from flask_potion.instances import Instances, RelationInstances
from flask_potion.reference import ResourceBound, ResourceReference
from flask_potion.schema import Schema, FieldSet


HTTP_METHODS = ('GET', 'PUT', 'POST', 'PATCH', 'DELETE')

def url_rule_to_uri_pattern(rule):
    # TODO convert from underscore to camelCase
    return re.sub(r'<(\w+:)?([^>]+)>', r'{\2}', rule)

def attribute_to_route_uri(s):
    return s.replace('_', '-')

def to_camel_case(s):
    return s[0].lower() + s.title().replace('_', '')[1:] if s else s

def _bind_schema(schema, resource):
    if isinstance(schema, ResourceBound):
        return schema.bind(resource)
    return schema


class Link(object):
    """
    This class is an container used by :class:`Route` to store method view functions and their schemas.

    Links are not bound to a specific resource and their schema, generated using :meth:`schema_factory` can vary depending on the resource.

    If ``view_func`` has an ``__annotations__`` attribute (a Python 3.x function annotation), the annotations
    will be used to generate the ``request_schema`` and ``response_schema``. The *return* annotation in this case
    is expected to be a :class:`schema.Schema` used for responses, and all other annotations are expected to be of type :class:`fields.Raw`
    and are combined into a :class:`schema.Fieldset`.

    :param callable view_func: view function
    :param str rel: relation
    :param str title: title of schema
    :param str description: description of schema
    :param routes.Route route: route this link belongs to
    :param str method: a HTTP request method name (upper case)
    :param schema.Schema schema: request schema
    :param schema.Schema response_schema: response schema
    :param bool format_response: whether the response should be converted using the response schema

    .. property:: relation

        A relation for the string, equal to ``rel`` if one was given.

    .. attribute:: request_schema

        request schema (not resource-bound)

    .. attribute:: response_schema

        response schema (not resource-bound)
    """
    def __init__(self,
                 view_func,
                 rel=None,
                 title=None,
                 description=None,
                 route=None,
                 method=None,
                 schema=None,
                 response_schema=None,
                 format_response=True):
        self.rel = rel
        self.title = title
        self.description = description
        self.route = route
        self.method = method
        self.view_func = view_func
        self.format_response = format_response

        annotations = getattr(view_func, '__annotations__', None)

        if isinstance(annotations, dict) and len(annotations):
            self.request_schema = FieldSet({name: field for name, field in annotations.items() if name != 'return'})
            self.response_schema = annotations.get('return', response_schema)
        else:
            self.request_schema = schema
            self.response_schema = response_schema

    @property
    def relation(self):
        if self.rel:
            return self.rel
        elif len(self.route.methods()) == 1:
            return self.route.attribute
        else:
            return "{}_{}".format(self.method, self.route.attribute)

    def schema_factory(self, resource):
        """
        Returns a link schema for a specific resource.
        """
        request_schema = _bind_schema(self.request_schema, resource)
        response_schema = _bind_schema(self.response_schema, resource)

        # NOTE "href" for rel="instances" MUST NOT be relative; others MAY BE relative
        schema = OrderedDict([
            ("rel", self.relation),
            ("href", url_rule_to_uri_pattern(self.route.rule_factory(resource, relative=False))),
            ("method", self.method)
        ])

        if self.title:
            schema["title"] = self.title
        if self.description:
            schema["description"] = self.description

        if request_schema:
            schema["schema"] = request_schema.request
        if response_schema:
            schema["targetSchema"] = response_schema.response
        return schema


def _method_decorator(method):
    def wrapper(self, *args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            return self._set_method_link(method, args[0], **kwargs)
        else:
            return lambda f: self._set_method_link(method, f, *args, **kwargs)

    wrapper.__name__ = method
    return wrapper


class Route(object):
    """

    .. decoratormethod:: METHOD(rule=None, attribute=None, rel=None, title=None, description=None, schema=None, response_schema=None, format_response=True)

        A decorator for registering the *METHOD* method handler of a route. Can be used with or without arguments and on both class and route instances. The ``rule`` and ``attribute``
        arguments are only available on the class.

        When used with an instance will add or replace the view function for the *METHOD* method of this :class:`Route` with the decorated function; otherwise instantiates a new :class:`Route` with the view function.

        This decorator is defined for the *GET*, *PUT*, *POST*, *PATCH* and *DELETE* methods.

        :param str rule: (class-only) route URI relative to the resource, defaults to ``/{attribute}``, replacing any ``'_'`` (underscore) in ``attribute`` with ``'-'`` (dash).
        :param str attribute: (class-only) attribute on the parent resource, used to identify the route internally; defaults to the attribute name of the decorated view function within the parent resource.
        :param str rel: relation of the method link to the resource
        :param str title: title of link schema
        :param str description: description of link schema
        :param schema.Schema schema: request schema
        :param schema.Schema schema: response schema
        :param bool format_response: whether the response should be converted using the response schema

    .. attribute:: request_schema

        Used to get and set the response schema for the most recently decorated request method view function

    .. attribute:: response_schema

        Used to get and set the response schema for the most recently decorated request method view function

    .. attribute:: method_links

        A dictionary mapping of method names (in upper case) to :class:`routes.Link` objects containing the method view functions.

    """
    def __init__(self, rule=None, attribute=None, format_response=True):
        self.rule = rule
        self.attribute = attribute
        self.format_response = format_response
        self.current_link = None
        self._method_links = {}

        for method in HTTP_METHODS:
            setattr(self, method, MethodType(_method_decorator(method), self))

    @classmethod
    def for_method(cls, method, func, rule=None, **kwargs):
        instance = cls(rule=rule,
                       attribute=kwargs.pop('attribute', None),
                       format_response=kwargs.pop('format_response', True))
        instance._set_method_link(method, func, **kwargs)
        return instance

    def _set_method_link(self, method, view_func, rel=None, title=None, description=None, schema=None, response_schema=None):
        link = Link(view_func,
                        rel=rel,
                        title=title,
                        description=description,
                        route=self,
                        method=method,
                        schema=schema,
                        response_schema=response_schema)
        self.current_link = link
        self._method_links[method] = link
        return self

    def __getattr__(self, name):
        return getattr(self.current_link, name)

    def __get__(self, obj, owner):
        if obj is None:
            return self
        return lambda *args, **kwargs: self.current_link.view_func.__call__(obj, *args, **kwargs)

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, repr(self.rule))

    @property
    def request_schema(self):
        return self.current_link.request_schema

    @request_schema.setter
    def request_schema(self, schema):
        self.current_link.request_schema = schema

    @property
    def response_schema(self):
        return self.current_link.response_schema

    @response_schema.setter
    def response_schema(self, schema):
        self.current_link.response_schema = schema

    def links(self):
        """
        Helper function that returns a list of all links within this route.
        """
        return self._method_links.values()

    def methods(self):
        """
        Helper function that returns a list of all methods supported by this route.
        """
        return [link.method for link in self.links()]

    def _rule_factory(self, resource, relative=False):
        rule = self.rule

        if rule is None:
            rule = '/{}'.format(attribute_to_route_uri(self.attribute))
        elif callable(rule):
            rule = rule(resource)

        if relative or resource.route_prefix is None:
            return rule[1:]

        return ''.join((resource.route_prefix, rule))

    def _view_factory(self, link, name, resource):
        request_schema = _bind_schema(link.request_schema, resource)
        response_schema = _bind_schema(link.response_schema, resource)
        view_func = link.view_func

        def view(*args, **kwargs):
            instance = resource()
            if isinstance(request_schema, (FieldSet, Instances)):
                kwargs.update(request_schema.parse_request(request))
            elif isinstance(request_schema, Schema):
                args += (request_schema.parse_request(request),)

            response = view_func(instance, *args, **kwargs)
            # TODO add 'described_by' header if response schema is a ToOne/ToMany/Instances field.
            if response_schema is None or not self.format_response:
                return response
            else:
                return response_schema.format_response(response)

        return view

    def rule_factory(self, resource, relative=False):
        """
        Returns a URL rule string for this route and resource.

        :param flask_potion.Resource resource:
        :param bool relative: whether the rule should be relative to ``resource.route_prefix``
        """
        return self._rule_factory(resource, relative)

    def view_factory(self, name, resource):
        """
        Returns a view function for all links within this route and resource.

        :param name: Flask view name
        :param flask_potion.Resource resource:
        """
        method_views = {}

        for name, link in self._method_links.items():
            method_views[name] = self._view_factory(link, name, resource)

        def view(*args, **kwargs):
            method = method_views.get(request.method)
            if method is None and request.method == 'HEAD':
                method = method_views['GET']
            return method(*args, **kwargs)
        return view


def _route_decorator(method):
    @classmethod
    def decorator(cls, *args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            return cls.for_method(method, args[0])
        else:
            return lambda f: cls.for_method(method, f, *args, **kwargs)
    if sys.version_info.major > 2:
        decorator.__name__ = method
    return decorator

for method in HTTP_METHODS:
    setattr(Route, method, _route_decorator(method))


class ItemRoute(Route):
    """
    This route can be used with :class:`flask_potion.ModelResource`. It is a simple extension over :class:`Route` with
    the following adjustments:

    - :meth:`rule_factory` is changed to prefix ``<{id_converter}:id>`` with any rule.
    - It changes the implementation of :meth:`view_factory` so that it passes the resolved resource item matching *id*
      as the first positional argument to the view function.
    """
    def _rule_factory(self, resource, relative=False):
        rule = self.rule
        id_matcher = '<{}:id>'.format(resource.meta.id_converter)

        if rule is None:
            rule = '/{}'.format(attribute_to_route_uri(self.attribute))
        elif callable(rule):
            rule = rule(resource)

        if relative or resource.route_prefix is None:
            return rule[1:]

        return ''.join((resource.route_prefix, '/', id_matcher, rule))

    def _view_factory(self, link, name, resource):
        original_view = super(ItemRoute, self)._view_factory(link, name, resource)
        def view(*args, **kwargs):
            id = kwargs.pop('id')  # Py2.7 -- could use (*args, id, **kwargs) otherwise
            item = resource.manager.read(id)
            return original_view(item, *args, **kwargs)
        return view


class RouteSet(object):
    """
    An abstract class for combining related routes into one, which can also be used as a route factory.
    """

    def routes(self):
        """
        :returns: an iterator over :class:`Route` objects
        """
        return ()


class ItemAttributeRoute(RouteSet):
    """

    :param fields.Raw cls_or_instance: a field class or instance
    :param str attribute: defaults to the field's ``attribute`` attribute
    :param str io: ``r``, ``w``, or ``rw`` - defaults to the field's ``io`` attribute
    """
    def __init__(self, cls_or_instance, io=None, attribute=None):
        self.field = _field_from_object(ItemAttributeRoute, cls_or_instance)
        self.attribute = attribute
        self.io = io

    def routes(self):
        io = self.io or self.field.io
        field = self.field
        route = ItemRoute(attribute=self.attribute)
        attribute = field.attribute or route.attribute

        if "r" in io:
            @route.GET(response_schema=field,
                       rel=to_camel_case('read_{}'.format(route.attribute)))
            def read_attribute(resource, item):
                return get_value(attribute, item, field.default)

        if "w" in io:
            @route.POST(schema=field,
                        response_schema=field,
                        rel=to_camel_case('update_{}'.format(route.attribute)))
            def update_attribute(resource, item, value):
                attribute = field.attribute or route.attribute
                item = resource.manager.update(item, {attribute: value})
                return get_value(attribute, item, field.default)

        yield route


class Relation(RouteSet, ResourceBound):
    """
    Used to define a relation to another :class:`ModelResource`.
    """
    def __init__(self, resource, backref=None, io="rw", attribute=None, **kwargs):
        self.reference = ResourceReference(resource)
        self.attribute = attribute
        self.backref = backref
        self.io = io

    @cached_property
    def target(self):
        return self.reference.resolve(self.resource)

    # FIXME can only be loaded after target is added to API
    def routes(self):
        io = self.io
        rule = '/{}'.format(attribute_to_route_uri(self.attribute))

        relation_route = ItemRoute(rule='{}/<{}:target_id>'.format(rule, self.target.meta.id_converter))
        relations_route = ItemRoute(rule=rule)

        if "r" in io:
            @relations_route.GET
            def relation_instances(resource, item, page, per_page):
                return resource.manager.relation_instances(item,
                                                           self.attribute,
                                                           self.target,
                                                           page,
                                                           per_page)

            relations_route.request_schema = FieldSet({
                "page": Integer(minimum=1, default=1),
                "per_page": Integer(minimum=1,
                                    default=20,  # FIXME use API reference
                                    maximum=50)
            })

            relations_route.response_schema = RelationInstances(self.target)

        if "w" in io:
            @relations_route.POST
            def relation_add(resource, item, target_item):
                resource.manager.relation_add(item, self.attribute, self.target, target_item)
                resource.manager.commit()
                return target_item

            relation_add.request_schema = ToOne(self.target)
            relation_add.response_schema = ToOne(self.target)

            @relation_route.DELETE
            def relation_remove(resource, item, target_id):
                target_item = self.target.manager.read(target_id)
                resource.manager.relation_remove(item, self.attribute, self.target, target_item)
                resource.manager.commit()
                return None, 204
            yield relation_route

        if io:
            yield relations_route


# class ItemMapAttribute(RouteSet):
#     """
#     Adds a route that includes the keys to a dictionary attribute and allows these to be read, set, and deleted
#     either individually or in bulk.
#
#     class Schema:
#         scores = fields.Object(field.Integer)
#
#     GET /item/:id/scores/:key/
#     PUT /item/:id/scores/:key/
#     DELETE /item/:id/scores/:key/
#
#     :param str mapping_attribute: Must map to an attribute that is of type :class:`str`
#     """
#
#     def __init__(self, cls_or_instance, io=None, **kwargs):
#         self.field = field = _field_from_object(self, cls_or_instance)
#         self.object_field = Object(field, **kwargs)
#         self.io = io
#
#     # @Route.GET(lambda r: '/<id:{}>/attribute/<key>'.format(r.meta.id_converter))
#     # def attribute(self, item, key):
#     #     pass
#
#     def routes(self):
#         field = self.field
#         object_field = self.object_field
#         attribute = field.attribute
#         rule = '/{}'.format(attribute_to_route_uri(attribute))
#         io = self.io or field.io
#
#         object = ItemRoute(rule=rule)
#         # object_property = ItemRoute(rule=rule)
#
#         @object.GET(response_schema=object_field)
#         def read(resource, item):
#             return get_value(attribute, item, {})
#
#         read.response_schema = object_field
#
#         @object.POST(schema=object_field)
#         def write(resource, item, value):
#             resource.manager.update(item, {attribute: value})
#             return value
#
#         # TODO PATCH object
#
#         # @object_property.GET(response_schema=field)
#         # def read_property(resource, item, key, value):
#         #     # TODO ensure key matches pattern.
#         #     pass
#         #
#         # @object_property.POST(schema=field)
#         # def write_property(resource, item, key, value):
#         #     # TODO ensure key matches pattern.
#         #     pass
#         #
#         # @object_property.DELETE
#         # def remove_property(resource, item, key):
#         #     # TODO ensure key matches pattern.
#         #     pass
#
#         yield object

#
# class ItemSetAttributeRoute(RouteSet):
#     """
#     GET /item/:id/values
#     POST /item/:id/value/:value_id
#     DELETE /item/:id/value/:value_id
#
#     The URL of each item must stay consistent when the order of the list changes.
#
#     Direct addressing is only supported when an `id_attribute` is given. This must be an attribute
#     that is unique and cannot be changed. Without it, only read-only mode can be supported.
#
#     Depending on the type of collection, sorting and filtering may or may not be supported.
#
#     """
#
#     def __init__(self, cls_or_instance, id_attribute=None, id_field=None, io="r"):
#         if "w" in io:
#             raise NotImplementedError("ItemListAttributeRoute only supports read-only mode for now.")
#
#         pass
#
#     # @ItemRoute.GET()
#     # def read_items(self):
#     #     pass
#     #
#     # @ItemRoute.GET()
#     # def read_item(self, item, position):
#     #     pass
#     #
#     # @read_item.POST()
#     # def add_item(self, item):
#     #     pass



from importlib import import_module
import inspect

import six
from flask import current_app, _app_ctx_stack


class ResourceReference(object):
    def __init__(self, value):
        self.value = value

    def resolve(self, binding=None):
        """
        Attempt to resolve the reference value and return the matching :class:`Resource`.
        """
        name = self.value

        if name == 'self':
            return binding

        from .resource import ModelResource
        if inspect.isclass(name) and issubclass(name, ModelResource):
            return name

        potion = None
        if binding and binding.api:
            potion = binding.api
        elif _app_ctx_stack.top is not None and hasattr(current_app, 'potion'):
            potion = current_app.potion
        if potion:
            if name in potion.resources:
                return potion.resources[name]

        try:
            if isinstance(name, six.string_types):
                module_name, class_name = name.rsplit('.', 1)
                module = import_module(module_name)
                return getattr(module, class_name)
        except ValueError:
            pass

        raise RuntimeError('Resource named "{}" is not registered with Potion'.format(name))

    def __repr__(self):
        return "<ResourceReference '{}'>".format(self.value)


class ResourceBound(object):
    resource = None

    def _on_bind(self, resource):
        pass

    def bind(self, resource):
        if self.resource is None:
            self.resource = resource
            self._on_bind(resource)
        elif self.resource != resource:
            return self.rebind(resource)
        return self

    def rebind(self, resource):
        raise NotImplementedError('{} is already bound to {}'
                                  ' and does not support rebinding to {}'.format(repr(self), self.resource, resource))


from __future__ import absolute_import

from operator import and_
from flask import current_app

import peewee as pw
try:
    from playhouse import postgres_ext
except ImportError:
    postgres_ext = False

from flask_potion import fields, signals
from flask_potion.backends import Manager
from flask_potion.exceptions import ItemNotFound, BackendConflict
from flask_potion.utils import get_value

PW_COMPARATOR_EXPRESSIONS = {
    '$eq': lambda column, value: column == value,
    '$ne': lambda column, value: column != value,
    '$in': lambda column, value: column << value,
    '$lt': lambda column, value: column < value,
    '$gt': lambda column, value: column > value,
    '$lte': lambda column, value: column <= value,
    '$gte': lambda column, value: column >= value,
    '$contains': lambda column, value: column.contains(value),
    '$startswith': lambda column, value: column.startswith(value.replace('%', '\\%')),
    '$endswith': lambda column, value: column.endswith(value.replace('%', '\\%')),
    '$istartswith': lambda column, value: column ** (value.replace('%', '\\%') + "%"),
    '$iendswith': lambda column, value: column ** ("%" + value.replace('%', '\\%'))
}


class PeeweeManager(Manager):
    """
    A manager for Peewee models.
    """
    supported_comparators = tuple(PW_COMPARATOR_EXPRESSIONS.keys())

    def __init__(self, resource, model):
        super(PeeweeManager, self).__init__(resource, model)

        meta = resource.meta
# TODO: Add support for composite primary keys.
        self.id_attribute = meta.get(
            'id_attribute', model._meta.primary_key.name)

        if 'id_field' in resource.meta:
            self.id_column = model._meta.fields[resource.meta.id_field]
        else:
            # TODO: Add support for composite primary keys.
            self.id_column = model._meta.primary_key

        if not hasattr(resource.Meta, 'name'):
            meta['name'] = model._meta.db_table.lower()

        fs = resource.schema
        include_fields = meta.get('include_fields', None)
        exclude_fields = meta.get('exclude_fields', None)
        read_only_fields = meta.get('read_only_fields', ())
        write_only_fields = meta.get('write_only_fields', ())
        pre_declared_fields = {f.attribute or k for k, f in fs.fields.items()}

        for name, column in model._meta.get_sorted_fields():
            if (include_fields and name in include_fields) or \
                    (exclude_fields and name not in exclude_fields) or \
                    not (include_fields or exclude_fields):
                if column.primary_key or name in model._meta.rel:
                    continue
                if name in pre_declared_fields:
                    continue

                args = ()
                kwargs = {}

                if isinstance(column, (pw.CharField, pw.TextField)):
                    field_class = fields.String
                    if hasattr(column, 'max_length') and column.max_length:
                        kwargs['max_length'] = column.max_length
                elif isinstance(column, pw.IntegerField):
                    field_class = fields.Integer
                elif isinstance(column, (pw.DecimalField, pw.FloatField)):
                    field_class = fields.Number
                elif isinstance(column, pw.BooleanField):
                    field_class = fields.Boolean
                elif isinstance(column, pw.DateField):
                    field_class = fields.Date
                elif isinstance(column, pw.DateTimeField):
                    field_class = fields.DateTime
                elif isinstance(column, pw.BlobField):
                    field_class = fields.raw
                elif isinstance(column, postgres_ext.ArrayField):
                    field_class = fields.Array
                    args = (fields.String,)
                elif postgres_ext and \
                        isinstance(column, postgres_ext.HStoreField):
                    field_class = fields.Object
                    args = (fields.String,)
                elif postgres_ext and \
                        isinstance(column, (postgres_ext.JSONField,
                                            postgres_ext.BinaryJSONField)):
                    field_class = fields.Raw
                    kwargs = {"schema": {}}
                else:
                    field_class = fields.String

                kwargs['nullable'] = column.null

                if column.default is not None:
                    kwargs['default'] = column.default

                io = "rw"
                if name in read_only_fields:
                    io = "r"
                elif name in write_only_fields:
                    io = "w"

                if not (column.null or column.default):
                    fs.required.add(name)

                fs.set(
                    name, field_class(*args, io=io, attribute=name, **kwargs))

    def _query(self):
        return self.model.select()

    def _where_expression(self, where):
        expressions = []

        for condition in where:
            column = getattr(self.model, condition.attribute)
            expressions.append(
                PW_COMPARATOR_EXPRESSIONS[condition.comparator.name](
                    column, condition.value))

        if len(expressions) == 1:
            return expressions[0]

        # TODO ranking by default with text-search.

        return and_(*expressions)

    def _order_by(self, sort):
        for attribute, reverse in sort:
            column = getattr(self.model, attribute)

            if reverse:
                yield column.desc()
            else:
                yield column.asc()

    def relation_instances(self, item, attribute, target_resource, page=None,
                           per_page=None):
        query = getattr(item, attribute)
        if page and per_page:
            return query.paginate(page, per_page)
        return query

    def relation_add(self, item, attribute, target_resource, target_item):
        signals.before_add_to_relation.send(
            self.resource, item=item, attribute=attribute, child=target_item)

        relation = getattr(item, attribute)
        if hasattr(relation, 'add'):
            relation.add(target_item)
        else:
            reverse_attribute = item._meta.reverse_rel[attribute].name
            setattr(target_item, reverse_attribute, item)
            target_item.save()

        signals.after_add_to_relation.send(
            self.resource, item=item, attribute=attribute, child=target_item)

    def relation_remove(self, item, attribute, target_resource, target_item):
        signals.before_remove_from_relation.send(
            self.resource, item=item, attribute=attribute, child=target_item)

        relation = getattr(item, attribute)
        if hasattr(relation, 'remove'):
            relation.remove(target_item)
        else:
            reverse_atribute = item._meta.reverse_rel[attribute].name
            setattr(target_item, reverse_atribute, None)
            target_item.save()

        signals.after_remove_from_relation.send(
            self.resource, item=item, attribute=attribute, child=target_item)

    def paginated_instances(self, page, per_page, where=None, sort=None):
        return self.instances(where, sort).paginate(page, per_page)

    def instances(self, where=None, sort=None):
        query = self._query()

        if where:
            query = query.where(self._where_expression(where))
        if sort:
            query = query.order_by(*self._order_by(sort))

        return query

    def first(self, where=None, sort=None):
        try:
            return self.instances(where, sort).first()
        except self.model.DoesNotExist:
            raise ItemNotFound(self.resource, where=where)

    def create(self, properties, commit=True):
        item = self.model()

        for key, value in properties.items():
            setattr(item, key, value)

        signals.before_create.send(
            self.resource, item=item)

        try:
            item.save()
        except pw.IntegrityError as e:
            print(e.args)
            if current_app.debug:
                raise BackendConflict(debug_info=e.args)
            raise BackendConflict()

        signals.after_create.send(self.resource, item=item)
        return item

    def read(self, id):
        try:
            return self.model.get(self.id_column == id)
        except self.model.DoesNotExist:
            raise ItemNotFound(self.resource, id=id)

    def update(self, item, changes, commit=True):
        actual_changes = {
            key: value for key, value in changes.items()
            if get_value(key, item, None) != value
        }

        signals.before_update.send(
            self.resource, item=item, changes=actual_changes)

        for key, value in changes.items():
            setattr(item, key, value)

        try:
            item.save()
        except pw.IntegrityError as e:
            if current_app.debug:
                raise BackendConflict(debug_info=e.args)
            raise BackendConflict()

        signals.after_update.send(
            self.resource, item=item, changes=actual_changes)
        return item

    def delete(self, item):
        signals.before_delete.send(
            self.resource, item=item)

        item.delete_instance()

        signals.after_delete.send(
            self.resource, item=item)



import re

import six
from .instances import Condition, COMPARATORS

from .schema import Schema
from .reference import ResourceBound
from .exceptions import ItemNotFound
from .utils import route_from, get_value


class Key(Schema, ResourceBound):

    def matcher_type(self):
        type_ = self.response['type']
        if isinstance(type_, six.string_types):
            return type_
        return type_[0]

    def rebind(self, resource):
        return self.__class__().bind(resource=resource)

    def schema(self):
        raise NotImplementedError()


class RefKey(Key):

    def matcher_type(self):
        return 'object'

    def schema(self):
        return {
            "type": "object",
            "properties": {
                "$ref": {
                    "type": "string",
                    "format": "uri",
                    "pattern": "^{}\/[^/]+$".format(re.escape(self.resource.route_prefix))
                }
                # TODO consider replacing with {$type: foo, $value: 123}
            },
            "additionalProperties": False
        }

    def _item_uri(self, resource, item):
        # return url_for('{}.instance'.format(self.resource.meta.id_attribute, item, None), get_value(self.resource.meta.id_attribute, item, None))
        return '{}/{}'.format(resource.route_prefix, get_value(resource.meta.id_attribute, item, None))

    def format(self, item):
        return {"$ref": self._item_uri(self.resource, item)}

    def convert(self, value):
        try:
            endpoint, args = route_from(value["$ref"], 'GET')
        except Exception as e:
            raise e
        # XXX verify endpoint is correct (it should be)
        # assert resource.endpoint == endpoint
        return self.resource.manager.read(args['id'])


class PropertyKey(Key):

    def __init__(self, property):
        self.property = property

    def rebind(self, resource):
        return self.__class__(self.property).bind(resource)

    def schema(self):
        return self.resource.schema.fields[self.property].response

    def format(self, item):
        return self.resource.schema.fields[self.property].output(self.property, item)

    def convert(self, value):
        return self.resource.manager.first(where=[Condition(self.property, COMPARATORS['$eq'], value)])


class PropertiesKey(Key):

    def __init__(self, *properties):
        self.properties = properties

    def matcher_type(self):
        return 'array'

    def rebind(self, resource):
        return self.__class__(*self.properties).bind(resource)

    def schema(self):
        return {
            "type": "array",
            "items": [self.resource.schema.fields[p].response for p in self.properties],
            "additionalItems": False
        }

    def format(self, item):
        return [self.resource.schema.fields[p].output(p, item) for p in self.properties]

    def convert(self, value):
        return self.resource.manager.first(where=[
            Condition(property, COMPARATORS['$eq'], value[i])
            for i, property in enumerate(self.properties)
        ])


class IDKey(Key):

    def _on_bind(self, resource):
        self.id_field = resource.meta.id_field_class(attribute=resource.meta.id_attribute)

    def schema(self):
        return self.id_field.response

    def format(self, item):
        return self.id_field.output(self.resource.meta.id_attribute, item)

    def convert(self, value):
        return self.resource.manager.read(self.id_field.convert(value))



from collections import OrderedDict

from flask import json
from werkzeug.utils import cached_property
from jsonschema import Draft4Validator, ValidationError, FormatChecker

from flask_potion.reference import ResourceBound
from flask_potion.utils import unpack
from flask_potion.exceptions import ValidationError as PotionValidationError, RequestMustBeJSON


class Schema(object):
    """
    The base class for all types with a schema in Potion. Has :attr:`response` and a :attr:`request` attributes
    for the schema to be used, respectively, for serializing and de-serializing.

    Any class inheriting from schema needs to implement :meth:`schema`.

    ..  attribute:: response

        JSON-schema used to represent data returned by the server.

    .. attribute:: request

        JSON-schema used for validation of data sent to the server.

    """

    def schema(self):
        """
        Abstract method returning the JSON schema used by both :attr:`response` and :attr:`request`.

        :return: a JSON-schema or a  tuple of JSON-schemas in the format ``(response_schema, request_schema)``
        """
        raise NotImplementedError()

    @cached_property
    def response(self):
        schema = self.schema()
        if isinstance(schema, tuple):
            return schema[0]
        return schema

    @cached_property
    def request(self):
        schema = self.schema()
        if isinstance(schema, tuple):
            return schema[1]
        return schema

    @cached_property
    def _validator(self):
        Draft4Validator.check_schema(self.request)
        return Draft4Validator(self.request, format_checker=FormatChecker())

    def format(self, value):
        """
        Formats a python object for JSON serialization. Noop by default.

        :param object value:
        :return:
        """
        return value

    def convert(self, instance):
        """
        Validates a deserialized JSON object against :attr:`request` and converts it into a python object.

        :param instance: JSON import
        :raises PotionValidationError: if validation failed
        """
        try:
            self._validator.validate(instance)
        except ValidationError as ve:
            errors = self._validator.iter_errors(instance)
            raise PotionValidationError(errors)
        return instance

    def parse_request(self, request):
        """
        Parses a Flask request object, validates it the against :attr:`request` and returns the converted request data.

        :param request: Flask request object
        :return:
        """
        data = request.json

        if not data and request.method in ('GET', 'HEAD'):
            data = dict(request.args)

        return self.convert(data)

    def format_response(self, response):
        """
        Takes a response value, which can be a ``data`` object or a tuple ``(data, code)`` or ``(data, code, headers)``
        and formats it using :meth:`format`.

        :param response: A response tuple.
        :return: A tuple in the form ``(data, code, headers)``
        """
        data, code, headers = unpack(response)
        # TODO omit formatting on certain codes.
        return self.format(data), code, headers


class SchemaImpl(Schema):
    def __init__(self, schema):
        self._schema = schema

    def schema(self):
        return self._schema


class FieldSet(Schema, ResourceBound):
    """
    A schema representation of a dictionary of :class:`fields.Raw` objects.

    Uses the fields' ``io`` attributes to determine whether they are read-only, write-only, or read-write.

    :param dict fields: a dictionary of :class:`fields.Raw` objects
    :param required_fields: a list or tuple of field names that are required during parsing
    """

    def __init__(self, fields, required_fields=None):
        self.fields = fields
        self.required = set(required_fields or ())

    def bind(self, resource):
        if self.resource is None:
            self.resource = resource
            self.fields = {
                key: field.bind(resource) if isinstance(field, ResourceBound) else field
                for key, field in self.fields.items()
            }
        elif self.resource != resource:
            return self.rebind(resource)
        return self

    def rebind(self, resource):
        return FieldSet(
            dict(self.fields),
            tuple(self.required)
        ).bind(resource)

    def set(self, key, field):
        if self.resource and isinstance(field, ResourceBound):
            field = field.bind(self.resource)
        self.fields[key] = field

    def _schema(self, patchable=False):
        response_schema = {
            "type": "object",
            "properties": OrderedDict((
                (key, field.response) for key, field in self.fields.items() if 'r' in field.io))
        }

        request_schema = {
            "type": "object",
            "additionalProperties": False,
            "properties": OrderedDict((
                (key, field.request) for key, field in self.fields.items() if 'w' in field.io))
        }

        for key, field in self.fields.items():
            if 'w' in field.io and not field.nullable and field.default is None:
                self.required.add(key)

        if not patchable and self.required:
            request_schema['required'] = list(self.required)

        return response_schema, request_schema


    def schema(self):
        return self._schema()

    @cached_property
    def patchable(self):
        return SchemaImpl(self._schema(True))

    def format(self, item):
        return OrderedDict((key, field.output(key, item)) for key, field in self.fields.items() if 'r' in field.io)

    def convert(self, instance, pre_resolved_properties=None, patchable=False, strict=False):
        """
        :param instance: JSON-object
        :param pre_resolved_properties: optional dictionary of properties that are already known
        :param bool patchable: when ``True`` does not check for required fields
        :param bool strict:
        :return:
        """
        result = dict(pre_resolved_properties) if pre_resolved_properties else {}

        if patchable:
            object_ = self.patchable.convert(instance)
        else:
            object_ = super(FieldSet, self).convert(instance)

        for key, field in self.fields.items():
            if 'w' not in field.io:
                continue

            # ignore fields that have been pre-resolved
            if key in result:
                continue

            value = None

            try:
                value = object_[key]
                value = field.convert(value, validate=False)
            except KeyError:
                if patchable:
                    continue

                if field.default is not None:
                    value = field.default
                elif field.nullable:
                    value = None
                elif key not in self.required and not strict:
                    value = None

            result[field.attribute or key] = value
        return result

    def parse_request(self, request):
        if request.method in ('POST', 'PATCH', 'PUT', 'DELETE'):
            if request.mimetype != 'application/json':
                raise RequestMustBeJSON()

        # TODO change to request.get_json() to catch invalid JSON
        data = request.json

        # FIXME raise error if request body is not JSON

        if not self.fields:
            return {}

        if not data and request.method in ('GET', 'HEAD'):
            data = {}

            for name, field in self.fields.items():
                try:
                    value = request.args[name]
                    # FIXME type conversion!
                    try:
                        data[name] = json.loads(value)
                    except TypeError:
                        data[name] = value
                except KeyError:
                    pass

        return self.convert(data, patchable=request.method == 'PATCH')

